==> sphere_connector-0.4.30.tar.gz: sphere_connector-0.4.30/sphere_connector/experiment.py <==
"""
The MIT License (MIT)
Copyright (c) 2014-2015 University of Bristol

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.
"""
import pandas as pd
import numpy as np
import os
import simplejson as json
import logging

from environmental import MOTION_SENSORS
from all_annotations import ALL_ANNOTATIONS
from code_timer import CodeTimer

from annotations import Annotations
from environmental import Environmental
from video import Video
from wearable import Wearable


class Experiment(object):
    def __init__(self, client, config, server_config, output_path, annotation_path):
        self.config = config
        self.file_stub = os.path.join(output_path, self.config.experiment_id)
        self.parts = ""

        self.client = client
        
        self.annotations = Annotations(server_config, self.client.db, annotation_path)
        self.environmental = Environmental(server_config, self.client.db)
        self.video = Video(server_config, self.client.db)
        self.wearable = Wearable(server_config, self.client.db)
        
        self.annotations.init_experiment(config)
        self.environmental.init_experiment(config)
        self.wearable.init_experiment(config)
        self.video.init_experiment(config)

        logging.debug(str(self))

        self.data = None
        self.df = None

    def __str__(self):
        if self.config.annotators:
            return 'Experiment ID: ' + self.config.experiment_id + ', annotators: ' + self.config.annotators
        else:
            return 'Experiment ID: ' + self.config.experiment_id

    def process(self, environmental=True, video=True, wearable=True, annotations=True):
        with CodeTimer('processing ...'):
            data = []
            all_true = True
            parts = "_"
            if environmental:
                data = data + self.environmental.get_data()
                parts += "E"
            else:
                all_true = False
            if video:
                data = data + self.video.get_data()
                parts += "V"
            else:
                all_true = False
            if wearable:
                data = data + self.wearable.get_data()
                parts += "W"
            else:
                all_true = False
            if annotations and self.config.annotations:
                data = data + self.annotations.get_data()
                parts += "A"
            else:
                all_true = False

            if not all_true:
                self.parts = parts

            self.df = pd.DataFrame(data)
            if 'datetime' in self.df:
                # Can only fail if the dataframe is empty
                self.df.set_index('datetime', inplace=True)

            if self.config.annotations:
                self.annotations.post_process(self.df)

            # self.df.sort(inplace=True)
            # self.df.sort(inplace=True, columns=['uid', 'trigger', 'tier'])
            # df['ann-steady'] = df['ann-trigger'].cumsum() + df['ann-trigger'] == -1
            # self.df.reset_index(inplace=True)

    def empty_column(self, resample_fill_method='ffill', resample_rule=None):
        ss = pd.Series([0], index=[self.config.experiment_start])
        se = pd.Series([0], index=[self.config.experiment_end])
        series = pd.concat([ss, se])
        return series.resample(resample_fill_method=resample_fill_method, resample_rule=resample_rule)

    def get_sensor_data(self, test_sensors, resample_fill_method='ffill', resample_rule=None, rescale=True,
                        fmt='series', compress=False):
        if fmt not in ['series', 'lists', 'json']:
            raise NotImplementedError('Unknown format ' + fmt)

        if fmt == 'json':
            yield '['

        first = True

        for sensor in test_sensors:
            if sensor not in self.df:
                logging.debug(sensor + ' not in data frame')
                series = self.empty_column(resample_fill_method, resample_rule)
                uid = ''
            else:
                dft = self.df[[sensor, 'uid']].copy().dropna()
                if dft.empty:
                    logging.debug(sensor + ' empty')
                    # print(dft.head())
                    series = self.empty_column(resample_fill_method, resample_rule)
                    uid = ''
                else:
                    for uid, grp in dft.groupby(['uid']):
                        grp.fillna("ffill", inplace=True)

                        series = grp[sensor]

                        if series.dtype == float:

                            if resample_fill_method and resample_rule:
                                logging.debug(sensor + ' before resampling ' + str(series.shape[0]))
                                ss = pd.Series([0], index=[self.config.experiment_start])
                                se = pd.Series([0], index=[self.config.experiment_end])

                                series = pd.concat([ss, series, se])
                                series = series.resample(resample_fill_method=resample_fill_method,
                                                         resample_rule=resample_rule)
                                # series.fillna(method="ffill", inplace=True)
                                logging.debug(sensor + ' after resampling ' + str(series.shape[0]))

                            if rescale:
                                series /= np.max(series)
                                series -= np.min(series)

                            if compress:
                                series = series.to_sparse()

                            if fmt == 'lists':
                                series = zip(series.index.to_pydatetime(), series.values)
                            if fmt == 'json':
                                series = zip(map(str, series.index.to_pydatetime()), series.values)

                        elif series.dtype == object:
                            # Maybe this is an array?
                            if fmt == 'series':
                                series = np.array([np.array(x) for x in series])
                            elif fmt == 'lists':
                                series = zip(series.index.to_pydatetime(), map(list, series.values))
                            else:
                                series = zip(map(str, series.index.to_pydatetime()), map(list, series.values))

            if fmt == 'series':
                # data.append((sensor, uid, series))
                yield (sensor, uid, series)
            elif fmt == 'lists':
                # data[sensor + '-' + uid] = {'data': series}
                yield (sensor + '-' + uid, {'data': series})
            elif fmt == 'json':
                if first:
                    first = False
                else:
                    yield ", "
                yield json.dumps({sensor + '-' + uid: {'data': series}})

        if fmt == 'json':
            yield ']'

    def get_motion_data(self, resample_fill_method='ffill', resample_rule=None, rescale=True,
                        fmt='series', compress=False):
        return self.get_sensor_data(MOTION_SENSORS, resample_fill_method, resample_rule, rescale, fmt, compress)

    def get_environmental_data(self, include_motion, resample_fill_method='ffill', resample_rule=None, rescale=True,
                               fmt='series', compress=False):
        test_sensors = self.environmental.sensors
        if not include_motion:
            test_sensors = set(test_sensors).difference(MOTION_SENSORS)
        return self.get_sensor_data(sorted(list(test_sensors)), resample_fill_method, resample_rule, rescale, fmt,
                                    compress)

    def get_wearable_data(self, resample_fill_method='ffill', resample_rule=None, rescale=False,
                          fmt='lists', compress=False):
        return self.get_sensor_data(self.wearable.sensors, resample_fill_method, resample_rule, rescale, fmt, compress)

    def get_video_data(self):
        data = []
        for sensor in enumerate(self.video.sensors):
            dft = self.df[[sensor, 'uid']].copy().dropna()
            dfg = dft.groupby('uid')
            for uid, grp in dfg:
                dat = np.array([np.array(x) for x in grp[sensor]])
                data.append((sensor, uid, dat))
        return data
        # return self.get_sensor_data(self.video.sensors)

    def get_annotations(self, annotator, tier):
        all_labels = ['None'] + ALL_ANNOTATIONS[tier]
        dft = self.df['ann-' + annotator + '-' + tier + '-discrete'].copy()
        dft = dft.sort_index()
        return dft, all_labels

    def update_annotations(self, doc, environmental=True, video=True, wearable=True, annotations=True):
        result = self.annotations.update(doc)
        if result['success']:
            self.process(environmental, video, wearable, annotations)
        # self.df.set_index('datetime', inplace=True)
        self.df.sort(inplace=True)
        self.df.sort(inplace=True, columns=['uid', 'trigger', 'tier'])
        return result

    def load(self):
        # self.df = pd.read_pickle(self.file_stub + ".pk")
        # self.df = pd.read_csv(self.file_stub + ".csv")
        df = pd.read_json(self.file_stub + self.parts + '.json')
        df.set_index('datetime', inplace=True)
        df.sort(inplace=True)
        df.sort(inplace=True, columns=['uid', 'trigger', 'tier'])
        self.df = df

    def save(self):
        messages = []
        self.df.sort(inplace=True)
        try:
            with open(self.file_stub + self.parts + '.csv', 'w+') as f:
                self.df.to_csv(f)
                messages.append("DataFrame saved to " + self.file_stub + self.parts + ".csv")
        except IOError as e:
            messages.append("Failed to save to " + self.file_stub + self.parts + ".csv")
            messages.append(e)

        try:
            self.df.to_pickle(self.file_stub + self.parts + '.pk')
            messages.append("DataFrame saved to " + self.file_stub + self.parts + ".pk")
        except IOError as e:
            messages.append("Failed to save to " + self.file_stub + self.parts + ".pk")
            messages.append(e)

        for message in messages:
            logging.info(message)
        return messages
        # df = self.df.reset_index()
        # df.to_json(self.file_stub + self.parts + '.json', date_format='iso')
        # print("DataFrame saved to " + self.file_stub + self.parts + ".json")

==> sphere_connector-0.5.4.tar.gz: sphere_connector-0.5.4/sphere_connector/experiment.py <==
"""
The MIT License (MIT)
Copyright (c) 2014-2015 University of Bristol

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.
"""
import pandas as pd
import numpy as np
import os
import simplejson as json
import logging

from environmental import MOTION_SENSORS
from all_annotations import ALL_ANNOTATIONS
from code_timer import CodeTimer

from annotations import Annotations
from environmental import Environmental
from video import Video
from wearable import Wearable


class Experiment(object):
    def __init__(self, client, experiment_config, basic_config):
        self.experiment_config = experiment_config
        self.file_stub = os.path.join(basic_config.output_path, experiment_config.experiment_id)
        self.parts = ""

        self.client = client
        
        self.annotations = Annotations(basic_config.mongo, self.client.db, basic_config.annotation_path)
        self.environmental = Environmental(basic_config.mongo, self.client.db)
        self.video = Video(basic_config.mongo, self.client.db)
        self.wearable = Wearable(basic_config.mongo, self.client.db)
        
        self.annotations.init_experiment(experiment_config)
        self.environmental.init_experiment(experiment_config)
        self.wearable.init_experiment(experiment_config)
        self.video.init_experiment(experiment_config)

        logging.debug(str(self))

        self.data = None
        self.df = None

    def __str__(self):
        if self.experiment_config.annotators:
            return 'Experiment ID: ' + self.experiment_config.experiment_id + \
                   ', annotators: ' + self.experiment_config.annotators
        else:
            return 'Experiment ID: ' + self.experiment_config.experiment_id

    def process(self, environmental=True, video=True, wearable=True, annotations=True):
        with CodeTimer('processing ...'):
            data = []
            all_true = True
            parts = "_"
            if environmental:
                data = data + self.environmental.get_data()
                parts += "E"
            else:
                all_true = False
            if video:
                data = data + self.video.get_data()
                parts += "V"
            else:
                all_true = False
            if wearable:
                data = data + self.wearable.get_data()
                parts += "W"
            else:
                all_true = False
            if annotations and self.experiment_config.annotations:
                data = data + self.annotations.get_data()
                parts += "A"
            else:
                all_true = False

            if not all_true:
                self.parts = parts

            self.df = pd.DataFrame(data)
            if 'datetime' in self.df:
                # Can only fail if the dataframe is empty
                self.df.set_index('datetime', inplace=True)

            if self.experiment_config.annotations:
                self.annotations.post_process(self.df)

            # self.df.sort(inplace=True)
            # self.df.sort(inplace=True, columns=['uid', 'trigger', 'tier'])
            # df['ann-steady'] = df['ann-trigger'].cumsum() + df['ann-trigger'] == -1
            # self.df.reset_index(inplace=True)

    def empty_column(self, resample=None):
        ss = pd.Series([0], index=[self.experiment_.experiment_start])
        se = pd.Series([0], index=[self.experiment_.experiment_end])
        series = pd.concat([ss, se])
        if resample:
            resample['how'] = None
            return series.resample(**resample)
        else:
            return series

    def get_sensor_data(self, test_sensors, resample=None, rescale=True, fmt='series', compress=False, threshold=None):
        if fmt not in ['series', 'lists', 'json']:
            raise NotImplementedError('Unknown format ' + fmt)

        if fmt == 'json':
            yield '['

        first = True

        for sensor in test_sensors:
            if sensor not in self.df:
                logging.debug(sensor + ' not in data frame')
                series = self.empty_column()
                uid = sensor.split('motion-')[1] if 'motion-' in sensor else ''
            else:
                dft = self.df[[sensor, 'uid']].copy().dropna()
                if dft.empty:
                    logging.debug(sensor + ' empty')
                    # print(dft.head())
                    series = self.empty_column()
                    uid = sensor.split('motion-')[1] if 'motion-' in sensor else ''
                else:
                    for uid, grp in dft.groupby(['uid']):
                        logging.debug(sensor + ': ' + uid + ' count=' + str(len(grp)))
                        grp.fillna("ffill", inplace=True)

                        series = grp[sensor]

                        if series.dtype == float:

                            if resample:
                                logging.debug(sensor + ' before resampling ' + str(series.shape[0]))
                                ss = pd.Series([0], index=[self.experiment_config.experiment_start])
                                se = pd.Series([0], index=[self.experiment_config.experiment_end])

                                series = pd.concat([ss, series, se])
                                series = series.resample(**resample)
                                # series.fillna(method="ffill", inplace=True)
                                logging.debug(sensor + ' after resampling ' + str(series.shape[0]))

                            if rescale:
                                series /= np.max(series)
                                series -= np.min(series)

                            if threshold:
                                series = series.clip(0, threshold, axis=0)

                            if compress:
                                series = series.to_sparse()

                            if fmt == 'lists':
                                series = zip(series.index.to_pydatetime(), series.values)
                            if fmt == 'json':
                                series = zip(map(str, series.index.to_pydatetime()), series.values)

                        elif series.dtype == object:
                            # Maybe this is an array?
                            if fmt == 'series':
                                series = np.array([np.array(x) for x in series])
                            elif fmt == 'lists':
                                series = zip(series.index.to_pydatetime(), map(list, series.values))
                            else:
                                series = zip(map(str, series.index.to_pydatetime()), map(list, series.values))

            if fmt == 'series':
                # data.append((sensor, uid, series))
                yield (sensor, uid, series)
            elif fmt == 'lists':
                # data[sensor + '-' + uid] = {'data': series}
                yield (sensor + '-' + uid, {'data': series})
            elif fmt == 'json':
                if first:
                    first = False
                else:
                    yield ", "
                yield json.dumps({sensor + '-' + uid: {'data': series}})

        if fmt == 'json':
            yield ']'

    def get_motion_data(self, resample, rescale=True, fmt='series', compress=False, threshold=1):
        return self.get_sensor_data(MOTION_SENSORS, resample, rescale, fmt, compress, threshold)

    def get_environmental_data(self, include_motion, resample, rescale=True, fmt='series', compress=False):
        test_sensors = self.environmental.sensors
        if not include_motion:
            test_sensors = set(test_sensors).difference(MOTION_SENSORS)
        return self.get_sensor_data(sorted(list(test_sensors)), resample, rescale, fmt, compress)

    def get_wearable_data(self, resample, rescale=False, fmt='lists', compress=False):
        return self.get_sensor_data(self.wearable.sensors, resample, rescale, fmt, compress)

    def get_video_data(self):
        data = []
        for sensor in enumerate(self.video.sensors):
            dft = self.df[[sensor, 'uid']].copy().dropna()
            dfg = dft.groupby('uid')
            for uid, grp in dfg:
                dat = np.array([np.array(x) for x in grp[sensor]])
                data.append((sensor, uid, dat))
        return data
        # return self.get_sensor_data(self.video.sensors)

    def get_annotations(self, annotator, tier):
        all_labels = ['None'] + ALL_ANNOTATIONS[tier]
        dft = self.df['ann-' + annotator + '-' + tier + '-discrete'].copy()
        dft = dft.sort_index()
        return dft, all_labels

    def update_annotations(self, doc, environmental=True, video=True, wearable=True, annotations=True):
        result = self.annotations.update(doc)
        if result['success']:
            self.process(environmental, video, wearable, annotations)
        # self.df.set_index('datetime', inplace=True)
        self.df.sort(inplace=True)
        self.df.sort(inplace=True, columns=['uid', 'trigger', 'tier'])
        return result

    def load(self):
        # self.df = pd.read_pickle(self.file_stub + ".pk")
        # self.df = pd.read_csv(self.file_stub + ".csv")
        df = pd.read_json(self.file_stub + self.parts + '.json')
        df.set_index('datetime', inplace=True)
        df.sort(inplace=True)
        df.sort(inplace=True, columns=['uid', 'trigger', 'tier'])
        self.df = df

    def save(self):
        messages = []
        self.df.sort(inplace=True)
        try:
            with open(self.file_stub + self.parts + '.csv', 'w+') as f:
                self.df.to_csv(f)
                messages.append("DataFrame saved to " + self.file_stub + self.parts + ".csv")
        except IOError as e:
            messages.append("Failed to save to " + self.file_stub + self.parts + ".csv")
            messages.append(e)

        try:
            self.df.to_pickle(self.file_stub + self.parts + '.pk')
            messages.append("DataFrame saved to " + self.file_stub + self.parts + ".pk")
        except IOError as e:
            messages.append("Failed to save to " + self.file_stub + self.parts + ".pk")
            messages.append(e)

        for message in messages:
            logging.info(message)
        return messages
        # df = self.df.reset_index()
        # df.to_json(self.file_stub + self.parts + '.json', date_format='iso')
        # print("DataFrame saved to " + self.file_stub + self.parts + ".json")
